/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2021, KR Soft s.r.o.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the KR2013ApS nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/
#ifndef ROBOT_STATUS_H
#define ROBOT_STATUS_H

#include <vector>
#include <array>
#include <algorithm>

#include <kr2/kord/robot_io.h>

namespace kr2 {
namespace kord {

//
// Frames
//
/// @brief Number of frames values available.
constexpr size_t num_frames_values{4 * 2}; // 2 references for each Frame

//
// Loads
//
/// @brief Number of loads available.
constexpr size_t num_loads{2};

// Encoder of temperature, sent in the form of uint16_t (a...b) = a...,b
inline uint16_t temperatureEncoder(const double& a_temp_decoded){
    uint16_t encoded_point = (a_temp_decoded * 10);
    return encoded_point;
}

struct JointSensorData{
    JointSensorData():
        jid_(UINT32_MAX),
        error_bits_(0),
        status_bits_(0),
        temperature_({0.0, 0.0, 0.0}),
        current_(0.0),
        voltage_(0.0),
        torque_(0.0)
    {

    }

    JointSensorData(unsigned int jid):
        jid_(jid),
        error_bits_(0),
        status_bits_(0),
        temperature_({0.0, 0.0, 0.0}),
        current_(0.0),
        voltage_(0.0),
        torque_(0.0)
    { }

    unsigned int jid_;
    unsigned int error_bits_;
    unsigned int status_bits_;
    std::array<double, 3>  temperature_;
    double current_;
    double voltage_;
    double torque_;
};

struct IOBoardData{
    IOBoardData():
        temperature_(0.0)
    {
    }
    double temperature_;
};

struct LoadData{
    LoadData():
        pose_({0.0, 0.0, 0.0, 0.0, 0.0, 0.0}),
        mass_(0.0),
        cog_({0.0, 0.0, 0.0}),
        inertia_({0.0, 0.0, 0.0, 0.0, 0.0, 0.0})
    {
    }
    uint8_t id_;
    std::vector<double> pose_;
    double mass_;
    std::vector<double> cog_;
    std::vector<double> inertia_;
};

struct FrameData{
    FrameData():
        pose_({0.0, 0.0, 0.0, 0.0, 0.0, 0.0})
    {
    }
    uint8_t id_;
    uint8_t ref_value_id_;
    std::vector<double> pose_;
};

struct RobotModel {
    std::vector<double> joint_position_;
    std::vector<double> joint_speed_;
    std::vector<double> joint_acceleration_;
    std::vector<double> joint_torque_;
    std::vector<double> tcp_model_;
    std::vector<double> tcp_model_fwk_;

    // Monitoring joints
    std::array<JointSensorData, 7> sensors_;

    //IOBoard state
    IOBoardData ioboard_;

    // Iterative update
    std::array<FrameData, num_frames_values> frames_;
    std::array<LoadData, num_loads> loads_;


     // State of robot ios
    RobotIO io_model_;

    std::vector<kr2rc_api::IOData::GPIOInt64> digital_inputs_;
    std::vector<kr2rc_api::IOData::GPIOInt64> digital_outputs_;

    std::vector<kr2rc_api::IOData::GPIOFloat> analog_inputs_;
    std::vector<kr2rc_api::IOData::GPIOFloat> analog_outputs_;

    void reset() {
        std::vector<double> v7d{0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
        std::vector<double> v6d{0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
        std::vector<double> v3d{0.0, 0.0, 0.0};

        joint_position_ = v7d;
        joint_speed_ = v7d;
        joint_acceleration_ = v7d;
        joint_torque_ = v7d;

        tcp_model_ = v6d;
        tcp_model_fwk_ = v6d;

        sensors_ = {JointSensorData(0),
                    JointSensorData(1),
                    JointSensorData(2),
                    JointSensorData(3),
                    JointSensorData(4),
                    JointSensorData(5),
                    JointSensorData(6) };
        
        ioboard_ = IOBoardData();
    }

    std::array<uint16_t, 7> getJointsTemperature(int a_type) const{
        std::array<uint16_t, 7> tmp;

        std::transform(sensors_.begin(), sensors_.end(), tmp.begin(),
            [a_type](const JointSensorData &js){
                return temperatureEncoder(js.temperature_.at(a_type));
            }
        );

        return tmp;
    }

    std::array<double, 7> getJointsTorques() const{
        std::array<double, 7> tmp;

        std::transform(sensors_.begin(), sensors_.end(), tmp.begin(),
            [](const JointSensorData &js){
                return js.torque_;
            }
        );

        return tmp;
    }

    uint16_t getIOBoardTemperature() const{
        return temperatureEncoder(ioboard_.temperature_);
    }


    std::array<double, 7> getJointsStatorVoltage() const
    {
        std::array<double, 7> tmp;

        std::transform(sensors_.begin(), sensors_.end(), tmp.begin(),
            [](const JointSensorData &js){
                return js.voltage_;
            }
        );

        return tmp;
    }

    std::array<double, 7> getJointsStatorCurrent() const
    {
        std::array<double, 7> tmp;

        std::transform(sensors_.begin(), sensors_.end(), tmp.begin(),
            [](const JointSensorData &js){
                return js.current_;
            }
        );

        return tmp;
    }

    std::array<unsigned int, 7> getJointsStatusBits() const
    {
        std::array<unsigned int, 7> tmp;

        std::transform(sensors_.begin(), sensors_.end(), tmp.begin(),
            [](const JointSensorData &js){
                return js.status_bits_;
            }
        );

        return tmp;
    }

    std::array<unsigned int, 7> getJointsErrorBits() const
    {
        std::array<unsigned int, 7> tmp;

        std::transform(sensors_.begin(), sensors_.end(), tmp.begin(),
            [](const JointSensorData &js){
                return js.error_bits_;
            }
        );

        return tmp;
    }
};

}  // kord
}  // kr2

#endif // ROBOT_STATUS_H
